#include <std_include.hpp>
#include "loader/component_loader.hpp"
#include "game/dvars.hpp"

#include "gsc/extension.hpp"

#include <utils/hook.hpp>

namespace player_movement {
namespace {
void __declspec(naked) pm_step_slide_move_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_bounces;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je bounce;

    cmp dword ptr [esp + 0x24], 0;
    jnz no_bounce;

   bounce:
    push 0x4E905B;
    ret;

   no_bounce:
    push 0x4E906F;
    ret;
  }
}

void __declspec(naked) pm_project_velocity_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_bouncesAllAngles;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je force_bounce;

    test ah, 0x5;
    jnp force_bounce;
    push 0x4761AF;
    ret;

   force_bounce:
    push 0x4761CF;
    ret;
  }
}

void __declspec(naked) pm_project_velocity2_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_bouncesAllAngles;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je force_bounce;

    fld1;
    jmp go_back;

   force_bounce:
    fldpi;

   go_back:
    fld [esp + 0x14 + 0x8];
    push 0x4761A4;
    ret;
  }
}

void __declspec(naked) pm_step_slide_move2_stub() {
  static float flt = 0.005f;
  __asm {
    push eax;
    mov eax, dvars::pm_bouncesAllAngles;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je force_bounce;

    fld ds:0x6D3AE4;
    jmp go_back;

   force_bounce:
    fld flt;

   go_back:
    push 0x4E9061;
    ret;
  }
}

game::gentity_s* weapon_rocket_launcher_fire_stub(
    game::gentity_s* ent, unsigned int weapon_index, float spread,
    game::weaponParms* wp, const float* gun_vel,
    game::lockonFireParms* lock_parms, bool magic_bullet) {
  auto* result = utils::hook::invoke<game::gentity_s*>(
      0x443F20, ent, weapon_index, spread, wp, gun_vel, lock_parms,
      magic_bullet);

  if (ent->client && dvars::pm_rocketJump->current.enabled) {
    const auto scale = dvars::pm_rocketJumpScale->current.value;
    ent->client->ps.velocity[0] += (0.0f - wp->forward[0]) * scale;
    ent->client->ps.velocity[1] += (0.0f - wp->forward[1]) * scale;
    ent->client->ps.velocity[2] += (0.0f - wp->forward[2]) * scale;
  }

  return result;
}

void cm_transformed_capsule_trace_stub(game::trace_t* results,
                                       const float* start, const float* end,
                                       const game::Bounds* bounds,
                                       const game::Bounds* capsule,
                                       int contents, const float* origin,
                                       const float* angles) {
  if (dvars::pm_playerCollision->current.enabled) {
    utils::hook::invoke<void>(0x47C7D0, results, start, end, bounds, capsule,
                              contents, origin, angles);
  }
}

void pm_trace_stub(game::pmove_t* pm, game::trace_t* results,
                   const float* start, const float* end,
                   const game::Bounds* bounds, int pass_entity_num,
                   int content_mask) {
  game::PM_trace(pm, results, start, end, bounds, pass_entity_num,
                 content_mask);
  // Allow the player to stand even when there is no headroom
  if (dvars::pm_elevators->current.enabled) {
    results->allsolid = false;
  }
}

void pm_player_trace_stub(game::pmove_t* pm, game::trace_t* results,
                          const float* start, const float* end,
                          const game::Bounds* bounds, int pass_entity_num,
                          int content_mask) {
  game::PM_playerTrace(pm, results, start, end, bounds, pass_entity_num,
                       content_mask);

  if (dvars::pm_elevators->current.enabled) {
    results->startsolid = false;
  }
}

void pm_crash_land_stub(const float* v, float scale, const float* result) {
  if (!dvars::pm_disableLandingSlowdown->current.enabled) {
    game::Vec3Scale(v, scale, result);
  }
}

void __declspec(naked) pm_accelerate_stub() {
  __asm {
    // Game's code
    sub esp, 0x20;

    push eax;
    mov eax, dvars::pm_csStrafe;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je attempt_to_force_movement;

    // Game's code
    test byte ptr [esi + 0xC], 8;
    jnz movement;

   go_back:
    push 0x64D87D;
    ret;

   attempt_to_force_movement:
    test byte ptr [esi + 0xC], 8;
    je movement;
    jmp go_back;

   movement:
    push 0x64D91A;
    ret;
  }
}

void __declspec(naked) cm_is_edge_walkable_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_terrainEdgeBounces;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je bounce;

    // Game's code
    lea eax, [eax + ecx * 2];
    add eax, ecx;
    push 0x5FAED5;
    ret;

   bounce:
    xor eax, eax;
    ret;
  }
}

void __declspec(naked) cm_is_edge_walkable2_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_terrainEdgeBounces;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je bounce;
    jmp go_back;

   bounce:
    mov cl, 0;

   go_back:
    pop ebp;
    mov byte ptr [edx + 0x2A], cl;
    pop ebx;
    add esp, 0x78;
    ret;
  }
}

void __declspec(naked) pm_step_slide_move3_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_doubleBounces;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je force_double_bounce;

    test ah, 0x5;
    jp go_back;

   force_double_bounce:
    push 0x4E904A;
    ret;

   go_back:
    push 0x4E90C8;
    ret;
  }
}

void __declspec(naked) pm_step_slide_move4_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_doubleBounces;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    // Code hook skipped
    add esp, 0xC;

    je force_double_bounce;

    push 0x4E8FB7;
    ret;

   force_double_bounce:
    push 0x4E8FD0;
    ret;
  }
}

void jump_clear_state_stub(game::playerState_s* ps) {
  if (!dvars::pm_doubleBounces->current.enabled) {
    game::Jump_ClearState(ps);
  }
}

void __declspec(naked) jump_check_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_bunnyHop;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    je auto_hop;

    // Game's code
    test dword ptr [ebx + 0x48], 0x400;
    push 0x4D25EF;
    ret;

   auto_hop:
    push 0x4D25FE;
    ret;
  }
}

// clang-format off
void __declspec(naked) pm_move_single_stub() {
  __asm {
    push eax;
    mov eax, dvars::pm_snapVector;
    cmp byte ptr [eax + 0x10], 1;
    pop eax;

    jnz skip;

    lea ebp, [ebp + 0x28]; // velocity
    push ebp;
    call game::Sys_SnapVector;
    add esp, 0x4;

   skip:

    // Game's code
    pop edi;
    pop esi;
    pop ebp;
    pop ebx;
    add esp, 0x90;
    ret;
  }
}
// clang-format on

void g_scr_is_sprinting(const game::scr_entref_t entref) {
  const auto* client = game::GetEntity(entref)->client;
  if (!client) {
    game::Scr_Error("IsSprinting can only be called on a player");
    return;
  }

  game::Scr_AddInt(game::PM_IsSprinting(&client->ps));
}
} // namespace

class component final : public component_interface {
public:
  static_assert(sizeof(game::weaponParms) == 0x48);
  static_assert(sizeof(game::lockonFireParms) == 0x18);

  void post_load() override {
    utils::hook(0x4E9054, pm_step_slide_move_stub, HOOK_JUMP)
        .install()
        ->quick(); // PM_StepSlideMove

    utils::hook(0x4761AA, pm_project_velocity_stub, HOOK_JUMP)
        .install()
        ->quick(); // PM_StepSlideMove
    utils::hook(0x47619E, pm_project_velocity2_stub, HOOK_JUMP)
        .install()
        ->quick(); // PM_StepSlideMove
    utils::hook(0x4E905B, pm_step_slide_move2_stub, HOOK_JUMP)
        .install()
        ->quick();

    utils::hook(0x4FA809, weapon_rocket_launcher_fire_stub, HOOK_CALL)
        .install()
        ->quick(); // FireWeapon

    utils::hook(0x4B6FC0, cm_transformed_capsule_trace_stub, HOOK_CALL)
        .install()
        ->quick(); // SV_ClipMoveToEntity
    utils::hook(0x57635F, cm_transformed_capsule_trace_stub, HOOK_CALL)
        .install()
        ->quick(); // CG_ClipMoveToEntity

    utils::hook(0x64F439, pm_trace_stub, HOOK_CALL)
        .install()
        ->quick(); // PM_CheckDuck
    utils::hook(0x651A60, pm_player_trace_stub, HOOK_CALL)
        .install()
        ->quick(); // PM_CorrectAllSolid
    utils::hook(0x651AF2, pm_player_trace_stub, HOOK_CALL)
        .install()
        ->quick(); // PM_CorrectAllSolid

    // Double bounces
    utils::hook(0x4E9045, pm_step_slide_move3_stub, HOOK_JUMP)
        .install()
        ->quick();
    utils::hook(0x4E8FAE, pm_step_slide_move4_stub, HOOK_JUMP)
        .install()
        ->quick();
    utils::hook(0x651D80, jump_clear_state_stub, HOOK_CALL)
        .install()
        ->quick(); // PM_GroundTrace

    // Bunnny hop
    utils::hook(0x4D25E8, jump_check_stub, HOOK_JUMP).install()->quick();
    // Enable / Disable jump landing punishment within PM_CrashLand
    utils::hook(0x64E571, pm_crash_land_stub, HOOK_CALL)
        .install()
        ->quick(); // Vec3Scale

    utils::hook(0x64D870, pm_accelerate_stub, HOOK_JUMP).install()->quick();

    // Implement terrain edge bounces
    utils::hook(0x5FAED0, cm_is_edge_walkable_stub, HOOK_JUMP)
        .install()
        ->quick();
    utils::hook(0x5FBB14, cm_is_edge_walkable2_stub, HOOK_JUMP)
        .install()
        ->quick();

    utils::hook(0x6530C3, pm_move_single_stub, HOOK_JUMP).install()->quick();

    gsc::add_method("IsSprinting", g_scr_is_sprinting);
    register_dvars();
  }

  static void register_dvars() {
    // clang-format off
    dvars::pm_bounces = game::Dvar_RegisterBool(
        "pm_bounces", false, game::DVAR_NONE, "CoD4 Bounces");
    dvars::pm_doubleBounces = game::Dvar_RegisterBool(
        "pm_doubleBounces", false, game::DVAR_NONE, "CoD4 Double Bounces");
    dvars::pm_bouncesAllAngles = game::Dvar_RegisterBool(
        "pm_bouncesAllAngles", false, game::DVAR_NONE, "Enable bouncing from all angles");
    dvars::pm_rocketJump = game::Dvar_RegisterBool(
        "pm_rocketJump", true, game::DVAR_NONE, "CoD4 rocket jumps");
    dvars::pm_csStrafe = game::Dvar_RegisterBool(
        "pm_csStrafe", false, game::DVAR_NONE, "Imitate CS Strafe movement");
    dvars::pm_rocketJumpScale = game::Dvar_RegisterFloat(
        "pm_rocketJumpScale", 64.0f, 0.0f, 1024.0f, game::DVAR_NONE, "");
    dvars::pm_playerCollision = game::Dvar_RegisterBool(
         "pm_playerCollision", true, game::DVAR_NONE, "Push intersecting players away from each other");
    dvars::pm_elevators = game::Dvar_RegisterBool(
         "pm_elevators", false, game::DVAR_NONE, "CoD4 elevators");
    dvars::pm_disableLandingSlowdown = game::Dvar_RegisterBool("pm_disableLandingSlowdown",
        false, game::DVAR_NONE, "Toggle landing slowdown");
    dvars::pm_bunnyHop = game::Dvar_RegisterBool("pm_bunnyHop",
        false, game::DVAR_NONE, "Constantly jump when holding space");
    dvars::pm_snapVector = game::Dvar_RegisterBool("pm_snapVector", false, game::DVAR_NONE, "Snap velocity");
    dvars::pm_terrainEdgeBounces = game::Dvar_RegisterBool(
        "pm_terrainEdgeBounces", false, game::DVAR_NONE, "Bounces on terrain edges");
    // clang-format on
  }
};
} // namespace player_movement

REGISTER_COMPONENT(player_movement::component)
